{   This file is a part of Witch7DRL, the roguelike game.
    Copyright (C) 2011 Konovod Andrey (kipar@front.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/
}
unit uClass;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graphics, Math, Dialogs, Forms, GL, uOpenGLGraph;

const
  USELOG = False;

  MAXX    = 1000;
  MAXY    = 1000;
  MINDIST = 5;
  DEFVIEWSCALE = 0.25;

  GENSOURCES = 500;

  GENOBSTR     = 75;
  GENOBSTRSIZE = 100;

  INITSPAWNRATE = 0.02;
  SPAWNINC      = 0.00005;

  MANACOST  = 0.0025;
  MANAREGEN = 0.001;
  MANARANGE = 150 * 150;
  DAMAGE    = 0.1;
  HEALPOWER = 0.05;

  MONSTERSPEED = MINDIST * 0.5;

  PLAYERSPEED = MINDIST;
  PLAYERARMOR = 0.5;

  DEMONSPEED = MINDIST;
  GOLEMSPEED = MINDIST * 0.8;

  PLAYERLIST  = 1;
  MONSTERLIST = 2;
  DEMONLIST   = 3;
  GOLEMLIST   = 4;

  TARGETMONSTERS = 1000;

  TRANSFORMTIME=200;
  TRANSFORMCOST=5;

type


  TFormFlag = (ffSeeSources, ffSeeSpells, ffHeals, ffBurns, ffDraw, ffFight, ffZombieFood, ffDemonFood, ffGolemFood);
  TFormFlags = set of TFormFlag;


  TGame = class;

  { TGameObject }

  TGameObject = class
    Game:   TGame;
    X, Y:   extended;
    Broken: boolean;
    destructor Destroy; override;
    procedure Draw; virtual;
    constructor Create(aGame: TGame; aX, aY: extended);
    procedure Process; virtual;
  end;

  TGameClass = class of TGameObject;

  { TUnit }

  TUnit = class(TGameObject)
    TargetX, TargetY, HP: extended;
    function Move: boolean;
    procedure Hit; virtual;
    procedure Die; virtual;
    procedure SetDirTo(ToX, ToY: extended);
    procedure Draw; override;
    procedure Process; override;
    constructor Create(aGame: TGame; aX, aY: extended); virtual;
    procedure Heal;
    function GetFlags: TFormFlags; virtual; abstract;
    function Armor: Extended; virtual;abstract;
    function Speed: Extended; virtual;abstract;
  end;

  TUnitClass = class of TUnit;

  { TMonster }

  TMonster = class(TUnit)
    Sidewalk: boolean;
    procedure Draw; override;
    constructor Create(aGame: TGame; aX, aY: extended); override;
    procedure Process; override;
    procedure Die; override;
    function GetFlags: TFormFlags; override;
    function Armor: Extended; override;
    function Speed: Extended; override;
  end;

  { TDemon }

  TDemon = class(TUnit)
    procedure Draw; override;
    procedure Process; override;
    function GetFlags: TFormFlags; override;
    function Armor: Extended; override;
    function Speed: Extended; override;
  end;

  { TGolem }

  TGolem = class(TUnit)
    WantPlayer: boolean;
    procedure Draw; override;
    procedure Process; override;
    function GetFlags: TFormFlags; override;
    function Armor: Extended; override;
    function Speed: Extended; override;
  end;



  { TObstracle }

  TObstracle = class(TGameObject)
    function CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean;
      virtual; abstract;
  end;

  { TRectObstracle }

  TRectObstracle = class(TObstracle)
    X2, Y2: extended;
    constructor Create(aGame: TGame; ax1, ay1, ax2, ay2: extended);
    procedure Draw; override;
    function CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean; override;
  end;

  { TTriObstracle }

  TTriObstracle = class(TObstracle)
    X2, Y2, X3, Y3: extended;
    constructor Create(aGame: TGame; ax1, ay1, ax2, ay2, ax3, ay3: extended);
    procedure Draw; override;
    function CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean; override;
  end;

  { TRoundObstracle }

  TRoundObstracle = class(TObstracle)
    R: extended;
    constructor Create(aGame: TGame; ax, ay, aR: extended);
    procedure Draw; override;
    function CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean; override;
  end;

  TElement      = (Fire, Water, Air, Earth);
  TElements     = set of TElement;
  TElementNames = array[TElement] of string;

  TLine = class;

  { TSource }

  TSource = class(TGameObject)
    Name:    string;
    MP:      extended;
    Element: TElement;
    Rechargeble: boolean;
    Lines:   TList;
    LastUsed: integer;
    procedure Draw; override;
    function UseMana(K: extended): boolean;
    constructor Create(aGame: TGame; ax, ay, aMP: extended; aElement: TElement);
    destructor Destroy; override;
    procedure Process; override;
    function Connected(Node: TSource): TLine;
  end;

  TExtendedPoint = packed record
    X, Y: extended;
  end;

  { TLine }

  TLine = class
    Game:      TGame;
    Start, Finish: TSource;
    LastAngle: real;
    Points:    array of TExtendedPoint;
    constructor Create(aGame: TGame; aStart: TSource);
    procedure AddPoint(aX, aY: extended; Angle: real);
    function Finalize(aFinish: TSource): boolean;
    procedure Draw(Style: TLineEffect);
    function OtherNode(S: TSource): TSource;
    procedure ToScratch;
  end;

  { TSpell }

  TSpell = class(TGameObject)
    Sources: TList;
    Lines:   TList;
    constructor Create(aGame: TGame);
    procedure AddLine(L: TLine); virtual;
    procedure Activate; virtual;
    procedure Draw; override;
    procedure Process; override;
    function UseMana(K: extended): boolean;
    destructor Destroy; override;
  end;

  { TDemoSpell }

  TDemoSpell = class(TSpell)
    procedure Process; override;
    constructor Create(aGame: TGame);
  end;

  { TChannelSpell }

  TChannelSpell = class(TSpell)
    procedure Process; override;
    procedure Draw; override;
    constructor Create(aGame: TGame; First, Second: TSource);
  end;

  { TPolyObstracle }

  TPolyObstracle = class(TObstracle)
    Points:  array of TExtendedPoint;
    TheRate: extended;
    function Inside(aX, aY: extended): boolean;
    procedure AddPoint(aX, aY: extended);
    procedure AddLine(L: TLine; FromPoint: TSource); virtual;
    function CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean; override;
    constructor Create(aGame: TGame);
    function RateSize: extended;
  end;

  { TAntiDaemonBarrier }

  TAntiDaemonBarrier = class(TPolyObstracle)
    Against:    TUnit;
    Disappear:  boolean;
    SavedLines: array of TLine;
    function CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean; override;
    procedure AddLine(L: TLine; FromPoint: TSource); override;
    constructor Create(aGame: TGame; Whom: TUnit);
    procedure Process; override;
  end;

  { TSummonSpell }

  TSummonSpell = class(TSpell)
    Summoned:    TUnit;
    SummonClass: TUnitClass;
    Barrier:     TAntiDaemonBarrier;
    constructor Create(aGame: TGame; aClass: TUnitClass);
    procedure Draw; override;
    procedure Activate; override;
    procedure Process; override;
  end;

{  TTeleportSpell = class(TSpell)
    Start, Finish: array[1..2] of TSource;
    constructor Create(aGame: TGame; aClass: TUnitClass; Line1, Line2: TLine);
    procedure Draw; override;
    procedure Activate;
    procedure Process; override;
  end;}


  TAreaSpell = class;

  { TSpecialObstracle }

  TSpecialObstracle = class(TPolyObstracle)
    Caller: TAreaSpell;
    constructor Create(aGame: TGame; aCaller: TAreaSpell);
    destructor Destroy; override;
    function CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean; override;
    procedure Process; override;
  end;

  { TAreaSpell }

  TAreaSpell = class(TSpell)
    MyArea: TSpecialObstracle;
    constructor Create(aGame: TGame);
    class function PreciseArea: boolean; virtual; abstract;
    function AreaColor(Index: extended): TOpenGLColor; virtual; abstract;
    procedure Activate; override;
    procedure Draw; override;
    destructor Destroy; override;
    function Callback(U: TUnit; aX, aY: extended): boolean; virtual; abstract;
  end;

  { TBarrier }

  TBarrier = class(TAreaSpell)
    class function PreciseArea: boolean; override;
    function Callback(U: TUnit; aX, aY: extended): boolean; override;
    procedure Process; override;
  end;


  { THealSpell }

  THealSpell = class(TAreaSpell)
    class function PreciseArea: boolean; override;
    function Callback(U: TUnit; aX, aY: extended): boolean; override;
    procedure Process; override;
    function AreaColor(Index: extended): TOpenGLColor; override;
  end;

  { TDamageSpell }

  TDamageSpell = class(TAreaSpell)
    class function PreciseArea: boolean; override;
    function Callback(U: TUnit; aX, aY: extended): boolean; override;
    procedure Process; override;
    function AreaColor(Index: extended): TOpenGLColor; override;
  end;


  TTransformEffect = (
    teNothing, teDamage, teSpeed,
    teWarrior, teZombie, teDemon,
    teFire, teWater, teEarth, teAir,
    teWarp
  );

  TTransRates = array[TTransformEffect] of Integer;

  TPassType = (NW, NE, SW, SE, None);

  TTransTable = array[TElement, TElement, TPassType, TPassType] of TTransformEffect;

  { TDontKnowSpell }

  TDontKnowSpell = class(TAreaSpell)
    North, West: TSource;
    SavedPass: TPassType;
    procedure Activate; override;
    class function PreciseArea: boolean; override;
    function Callback(U: TUnit; aX, aY: Extended): boolean; override;
    function GetPassType(X1, Y1, X2, Y2: Extended): TPassType;
    procedure Process; override;
    procedure Draw; override;
    procedure AddLine(L: TLine); override;
    procedure DoTransformEffect(E: TTransformEffect);
  end;




  TPattern = class
    function Recognize(Game: TGame; Line: TLine): TSpell; virtual; abstract;
  end;

  { TCirclePattern }

  TCirclePattern = class(TPattern)
    function Recognize(Game: TGame; Line: TLine): TSpell; override;
  end;

  TLinkCheckResult = (PartialOK, FullOK, IsFail, LastOK);

  TPolygonResult = (Convex, Polygon, NotPoly, CantCheck);
  TPolygonResults = set of TPolygonResult;

  { TObjectList }

  TObjectList = class
    Data: array of TObject;
    constructor Create;
    destructor Destroy; override;
    procedure PushObject(S: TObject);
    procedure PopObject;
    function FindSource(S: TObject): boolean;
    function PrevNode(N: integer): TObject;

    function X(I: integer): extended;
    function Y(I: integer): extended;
    function IsPolygon: TPolygonResult;
    function LogList: string;

  end;

  { TSequencePattern }

  TSequencePattern = class(TPattern)
    __FirstNode: TSource;
    __FinalLines, __FinalNodes: TObjectList;
    __Index:     integer;
    function MaxLen: integer; virtual; abstract;
    function CanRepeat: boolean; virtual;
    function NeedPoly: TPolygonResults;virtual;
    function RecognizePart(CurIndex: integer; FromNode: TSource;
      FirstNode: TSource; Exclude, FinalLines, FinalNodes: TObjectList): TSource;
    function CheckLink(Index: integer; Line: TLine; Node: TSource;
      FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult; virtual;
    function FullRecognize(aLine: TLine; FinalLines, FinalNodes: TObjectList): TSource;
  end;

  { TFixedSequencePattern }

  TSpellStyle = (Channel, Dontknow, DoDamage, Heal, Teleport);

  TFixedSequencePattern = class(TSequencePattern)
    Sequence:  array of TElement;
    MustClose: boolean;
    Style:     TSpellStyle;
    function CanRepeat: boolean; override;
    function MaxLen: integer; override;
    function NeedPoly: TPolygonResults;override;
    function Recognize(Game: TGame; Line: TLine): TSpell; override;
    constructor Create(aSeq: string; aStyle: TSpellStyle; aMustClose: boolean = False);
    function CheckLink(Index: integer; Line: TLine; Node: TSource;
      FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult; override;
  end;

  { TSetSequencePattern }

  TSetSequencePattern = class(TSequencePattern)
    FCanRepeat: boolean;
    Elements:   TElements;
    Style:      TSpellStyle;
    function CanRepeat: boolean; override;
    function MaxLen: integer; override;
    function NeedPoly: TPolygonResults;override;
    function Recognize(Game: TGame; Line: TLine): TSpell; override;
    constructor CreateOneElem(Elem: TElement; aStyle: TSpellStyle);
    constructor CreateManyElems(Elems: TElements; aStyle: TSpellStyle);
    function CheckLink(Index: integer; Line: TLine; Node: TSource;
      FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult; override;
  end;

  { THexagram }

  THexagram = class(TSequencePattern)
    function CanRepeat: boolean; override;
    function MaxLen: integer; override;
    function NeedPoly: TPolygonResults;override;
    function Recognize(Game: TGame; Line: TLine): TSpell; override;
    function CheckLink(Index: integer; Line: TLine; Node: TSource;
      FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult; override;
  end;

  { TPentagram }

  TPentagram = class(TSequencePattern)
    function CanRepeat: boolean; override;
    function MaxLen: integer; override;
    function NeedPoly: TPolygonResults;override;
    function Recognize(Game: TGame; Line: TLine): TSpell; override;
    function CheckLink(Index: integer; Line: TLine; Node: TSource;
      FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult; override;
  end;


  TPlayerForm = (pfNormal, pfWarrior, pfZombie, pfDemon, pfSpeed, pfElemental);

  { TPlayer }

  TPlayer = class(TUnit)
    Inventory: array[TElement] of integer;
    CurLine:   TLine;
    DrawMode:  boolean;

    CurForm: TPlayerForm;
    CurElement: TElement;
    CurTimeout: Integer;
    CurSource: TDontKnowSpell;

    procedure Draw; override;
    procedure InsertSource(E: TElement);
    procedure Die; override;
    constructor Create(aGame: TGame; aX, aY: extended); override;
    procedure Process; override;
    procedure CheckLines;
    procedure Hit; override;
    function GetFlags: TFormFlags; override;
    procedure SetForm(F: TPlayerForm; Time: Integer; Source: TDontKnowSpell; Elem: TElement = Fire);
    procedure ReturnForm;
    function Armor: Extended; override;
    function Speed: Extended; override;
  end;


  { TGame }

  TGame = class
    Objects:   TList;
    Player:    TPlayer;
    Scratches: TList;
    Patterns:  TList;
    CurSpawnRate: extended;
    VIEWSCALE: extended;
    TRANSTABLE: TTransTable;
    MonstersRemains: integer;
    WasWarped: Boolean;
    procedure InitPatterns;
    function CheckPatterns(L: TLine): TSpell;
    procedure ProcessTime;
    procedure Clear;

    procedure ClearTransTable;
    procedure GenTransTable;

    procedure LoadLevel(Filename: string);
    procedure GenLevel;
    procedure Draw;
    function CheckObstracles(Who: TUnit; var X: extended; var Y: extended): boolean;
    constructor Create;
    function NearObject(X, Y: Extended; aClass: TGameClass;
      Exclude: TGameObject = nil): TGameObject;
    function NearUnit(X, Y: Extended; ForFlag: TFormFlag;
      Exclude: TGameObject = nil): TUnit;
    function NearestObject(X, Y: Extended; aClass: TGameClass;
      Exclude: TGameObject = nil): TGameObject;
    function NearestUnit(X, Y: Extended; ForFlag: TFormFlag;
      Exclude: TGameObject = nil): TUnit;

  end;

const
  ElementNames: TElementNames = ('F', 'W', 'A', 'E');

  TransformRates: TTransRates =
    (
      200,  //    teNothing,  OK!
      50,   //    teDamage,   OK!
      20,   //    teSpeed,    OK!
      20,   //    teWarrior,  OK!
      20,   //    teZombie,   OK
      20,   //    teDemon,    OK?
      20,   //    teFire,
      20,   //    teWater,
      20,   //    teEarth,
      20,   //    teAir,
      5     //    teWarp
    );



function PointsNear(X1, Y1, X2, Y2: extended): boolean;
function IntersectLines(X1, Y1, X2, Y2, X3, Y3, X4, Y4: extended): boolean;
function InTria(X1, Y1, X2, Y2, X3, Y3: extended; X, Y: extended): boolean;
function RandomElement: TElement;

function ElementColor(E: TElement): TOpenGLColor;
function ElementCoreColor(E: TElement): TOpenGLColor;


var
  THEGAME: TGame;

implementation

uses uMain;

function PointsNear(X1, Y1, X2, Y2: extended): boolean;
begin
  Result := Sqr(X1 - X2) + Sqr(Y1 - Y2) < Sqr(MINDIST);
end;

{Эта функция украдена из интернета}
function IntersectLines(X1, Y1, X2, Y2, X3, Y3, X4, Y4: extended): boolean;

  function Q(ax, ay, bx, by, tx, ty: extended): extended;
  begin
    Result := tx * (by - ay) + ty * (ax - bx) + ay * bx - ax * by;
  end;

begin
  Result :=
    (Q(X1, Y1, X2, Y2, X3, Y3) * Q(X1, Y1, X2, Y2, X4, Y4) < 0) and
    (Q(X3, Y3, X4, Y4, X1, Y1) * Q(X3, Y3, X4, Y4, X2, Y2) < 0);
end;

{Эта функция украдена из интернета}
function InTria(X1, Y1, X2, Y2, X3, Y3: extended; X, Y: extended): boolean;
var
  f1, f2, f3: boolean;
begin
  f1     := (((x - x1) * (y1 - y2) + (y - y1) * (x2 - x1)) > 0);
  f2     := (((x - x2) * (y2 - y3) + (y - y2) * (x3 - x2)) > 0);
  f3     := (((x - x3) * (y3 - y1) + (y - y3) * (x1 - x3)) > 0);
  Result := (f1 = f2) and (f2 = f3);
end;

function RandomElement: TElement;
begin
  Result := TElement(Random(Ord(High(TElement)) + 1));
end;

function ElementColor(E: TElement): TOpenGLColor;
begin
  case E of
    Fire: Result  := RED;
    Water: Result := BLUE;
    Air: Result   := SKYBLUE;
    Earth: Result := GREEN;
  end;
end;

function ElementCoreColor(E: TElement): TOpenGLColor;
begin
  case E of
    Fire: Result  := YELLOW;
    Water: Result := SKYBLUE;
    Air: Result   := WHITE;
    Earth: Result := CLR(0.3, 0.3, 0, 1);
  end;
end;

{ TDontKnowSpell }

procedure TDontKnowSpell.Activate;
var
  MinX, MinY: Extended;
  I: Integer;
  Src: TSource;

begin
  inherited Activate;
  SavedPass := None;
  MinY := MAXY;
  North := nil;
  for I := 0 to Sources.Count - 1 do
    with TSource(Sources[I])do
      if Y < MinY then
      begin
        North := TSource(Sources[I]);
        MinY := Y;
      end;
  assert(North<>nil);
  MinX := MAXX;
  West := nil;
  for I := 0 to Lines.Count - 1 do
    with TLine(Lines[I]) do
    begin
      Src := OtherNode(North);
      if (Src <> nil) and (Src.X < MinX) then
      begin
        West := Src;
        MinX := Src.X;
      end;
    end;
  assert(West<>nil)
end;

class function TDontKnowSpell.PreciseArea: boolean;
begin
  Result := False;
end;

function TDontKnowSpell.Callback(U: TUnit; aX, aY: extended): boolean;
var
  P: TPassType;
  WasIn, NowIn: Boolean;
begin
  Result := True;
  if (U = Game.Player)  then
  begin
    WasIn := MyArea.Inside(U.X, U.Y);
    NowIn := MyArea.Inside(aX, aY);
    if WasIn = NowIn then exit;
    P := GetPassType(U.X, U.Y, aX, aY);
    if NowIn then
      SavedPass := P
    else if SavedPass <> None then
    begin
      DoTransformEffect(Game.TRANSTABLE[North.Element, West.Element, SavedPass, P]);
      SavedPass := None;
    end;
  end;
end;

function TDontKnowSpell.GetPassType(X1, Y1, X2, Y2: Extended): TPassType;
var
  I: Integer;
begin
  Result := None;
  for I := 0 to Lines.Count-1 do
    with TLine(Lines[I])do
      if IntersectLines(Start.X, Start.Y, Finish.X, Finish.Y, X1,Y1,X2,Y2)then
      begin
        if OtherNode(North) = nil then
        begin
          if OtherNode(West) = nil then
            Result := SE
          else
            Result := SW
        end
        else
        begin
          if OtherNode(West) = nil then
            Result := NE
          else
            Result := NW
        end;
        exit
      end;
end;

procedure TDontKnowSpell.Process;
var
  I: Integer;
begin
  inherited Process;
  if not UseMana(0) then
  begin
    for I := 0 to Sources.Count-1 do
      with TSource(Sources[I]) do
      begin
        Rechargeble := False;
        MP := -1;
      end;
    Free;
  end;
end;

procedure TDontKnowSpell.Draw;
begin

end;

procedure TDontKnowSpell.AddLine(L: TLine);
begin
  inherited;
{  Lines.Add(L);
  if Sources.IndexOf(L.Start) < 0 then
    Sources.Add(L.Start);
  if Sources.IndexOf(L.Finish) < 0 then
    Sources.Add(L.Finish);}
end;

procedure TDontKnowSpell.DoTransformEffect(E: TTransformEffect);
begin
  UseMana(1);
  case E of
  teDamage: Game.Player.Hp := Game.Player.Hp / 2;
  teSpeed: Game.Player.SetForm(pfSpeed, TRANSFORMTIME, Self);
  teWarrior: Game.Player.SetForm(pfWarrior, TRANSFORMTIME, Self);
  teZombie: Game.Player.SetForm(pfZombie, TRANSFORMTIME, Self);
  teDemon: Game.Player.SetForm(pfDemon, TRANSFORMTIME, Self);
  teFire: Game.Player.SetForm(pfElemental, 1, Self, Fire);
  teWater: Game.Player.SetForm(pfElemental, 1, Self, Water);
  teEarth: Game.Player.SetForm(pfElemental, 1, Self, Earth);
  teAir: Game.Player.SetForm(pfElemental, 1, Self, Air);
  teWarp: if not Game.WasWarped then
    begin
      Game.WasWarped := True;
      Game.GenLevel;
    end;
  end;
end;

{ TAntiDaemonBarrier }

function TAntiDaemonBarrier.CanStep(Who: TUnit; var aX: extended;
  var aY: extended): boolean;
begin
  Result := (Who <> Against) or Inside(aX, aY);
end;

{procedure TAntiDaemonBarrier.Draw;
var
  I, J: integer;
const
  NN = 1;
begin
  glBegin(GL_POINTS);
  for I := 1 to MAXX div NN do
    for J := 1 to MAXY div NN do
      if Inside(I * NN, J * NN) then
      begin
        glVertex2f(I * NN, J * NN);
      end;
  glEnd;
end;}{TODO - приспособить для полей?}

procedure TAntiDaemonBarrier.AddLine(L: TLine; FromPoint: TSource);
begin
  inherited AddLine(L, FromPoint);
  SetLength(SavedLines, Length(SavedLines) + 1);
  SavedLines[High(SavedLines)] := L;
end;

constructor TAntiDaemonBarrier.Create(aGame: TGame; Whom: TUnit);
begin
  inherited Create(aGame);
  Against   := Whom;
  Disappear := False;
  SetLength(SavedLines, 0);
end;

procedure TAntiDaemonBarrier.Process;
begin
  inherited Process;
  if Disappear then
    Free;
end;

{ TRoundObstracle }

constructor TRoundObstracle.Create(aGame: TGame; ax, ay, aR: extended);
begin
  inherited Create(aGame, ax, ay);
  R := ar;
end;

procedure TRoundObstracle.Draw;
begin
  glPushMatrix;
  glTranslatef(X, Y, 0);
  Circle(R, TRANSPARENT, BLACK, BLACK);
  glPopMatrix;
end;

function TRoundObstracle.CanStep(Who: TUnit; var aX: extended;
  var aY: extended): boolean;
begin
  Result := sqr(X - aX) + sqr(Y - aY) > sqr(R);
end;

{ TTriObstracle }

constructor TTriObstracle.Create(aGame: TGame; ax1, ay1, ax2, ay2, ax3, ay3: extended);
begin
  inherited Create(aGame, ax1, ay1);
  X2 := ax2;
  Y2 := ay2;
  X3 := ax3;
  Y3 := ay3;
end;

procedure TTriObstracle.Draw;
begin
  glEnable(GL_TEXTURE_2D);
  SetColor(GRAY);
  glBindTexture(GL_TEXTURE_2D, TEX_WALL);
  glBegin(GL_TRIANGLES);
  glTexCoord2f(0, 0);
  glVertex2f(X, Y);
  glTexCoord2f((X2 - X) / MINDIST, (Y2 - Y) / MINDIST);
  glVertex2f(X2, Y2);
  glTexCoord2f((X3 - X) / MINDIST, (Y3 - Y) / MINDIST);
  glVertex2f(X3, Y3);
  glEnd;
  glDisable(GL_TEXTURE_2D);
end;

function TTriObstracle.CanStep(Who: TUnit; var aX: extended;
  var aY: extended): boolean;
begin
  //Result := not InTria(X, Y, X2, Y2, X3, Y3, aX, aY);
  Result := not (InTria(X, Y, X2, Y2, X3, Y3, aX, aY) or
    IntersectLines(Who.X, Who.Y, aX, aY, X, Y, X2, Y2) or
    IntersectLines(Who.X, Who.Y, aX, aY, X, Y, X3, Y3) or
    IntersectLines(Who.X, Who.Y, aX, aY, X3, Y3, X2, Y2));
end;

{ TDamageSpell }

class function TDamageSpell.PreciseArea: boolean;
begin
  Result := False;
end;

function TDamageSpell.Callback(U: TUnit; aX, aY: extended): boolean;
begin
  Result := True;
  //  if U.InheritsFrom(TMonster) then U.Hit;
end;

procedure TDamageSpell.Process;
var
  I: integer;
begin
  inherited Process;
  if UseMana(MyArea.TheRate / MANARANGE) then
  begin
    for I := 0 to Game.Objects.Count - 1 do
      if TGameObject(Game.Objects[I]).InheritsFrom(TUnit) and
        (ffBurns in TUnit(Game.Objects[I]).GetFlags)then
        with TUnit(Game.Objects[I]) do
          if MyArea.Inside(X, Y) then
          begin
            UseMana(1);
            Hit;
          end;
  end
  else
    Free;
end;

function TDamageSpell.AreaColor(Index: extended): TOpenGLColor;
begin
  Result   := MixColor(RED, YELLOW, Index);
  Result.A := 0.4;
end;

{ THealSpell }

class function THealSpell.PreciseArea: boolean;
begin
  Result := False;
end;

function THealSpell.Callback(U: TUnit; aX, aY: extended): boolean;
begin
  Result := True;
end;

procedure THealSpell.Process;
var
  I: integer;
begin
  inherited Process;
  if UseMana(MyArea.TheRate / MANARANGE) then
  begin
    for I := 0 to Game.Objects.Count - 1 do
      if TGameObject(Game.Objects[I]).InheritsFrom(TUnit) and
        (ffHeals in TUnit(Game.Objects[I]).GetFlags)then
        with TUnit(Game.Objects[I]) do
          if MyArea.Inside(X, Y) then
          begin
            Heal;
            UseMana(3);
          end;
  end
  else
    Free;
end;

function THealSpell.AreaColor(Index: extended): TOpenGLColor;
begin
  Result   := MixColor(BLUE, GREEN, Index);
  Result.A := 0.2;
end;

{ TPentagram }

function TPentagram.CanRepeat: boolean;
begin
  Result := False;
end;

function TPentagram.MaxLen: integer;
begin
  Result := 5;
end;

function TPentagram.NeedPoly: TPolygonResults;
begin
  Result := inherited NeedPoly;
end;

function TPentagram.Recognize(Game: TGame; Line: TLine): TSpell;
var
  I: integer;
  N: TSource;
  Nodes, Lines: TObjectList;
begin
  Nodes := TObjectList.Create;
  Lines := TObjectList.Create;
  try
    N := FullRecognize(Line, Lines, Nodes);
    if (N <> nil) and (N = Line.Finish) then
    begin
      Result := TSummonSpell.Create(Game, TDemon);
      for I := 1 to Length(Nodes.Data) - 1 do
        Result.Sources.Add(Nodes.Data[I]);
      for I := 0 to Length(Lines.Data) - 1 do
        Result.AddLine(TLine(Lines.Data[I]));
      Result.Activate;
    end
    else
      Result := nil;
  finally
    Nodes.Free;
    Lines.Free;
  end;
end;

function TPentagram.CheckLink(Index: integer; Line: TLine; Node: TSource;
  FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;

  function Inters(I1, I2, J1, J2: integer): boolean;
  begin
    with FinalNodes do
      Result := IntersectLines(X(I1), Y(I1), X(I2), Y(I2), X(J1), Y(J1), X(J2), Y(J2));
  end;

begin
  Result := IsFail;
  {TODO: Инкрементальная проверка пересечений}
  if Index = 4 then
  begin
    if not Inters(1, 2, 3, 4) then
      exit;
    if not Inters(1, 2, 5, 4) then
      exit;
    if not Inters(1, 5, 3, 4) then
      exit;
    if not Inters(1, 5, 3, 2) then
      exit;
    if not Inters(2, 3, 4, 5) then
      exit;
  end;
  Result := inherited;
end;

{ TSummonSpell }

constructor TSummonSpell.Create(aGame: TGame; aClass: TUnitClass);
begin
  inherited Create(aGame);
  SummonClass := aClass;
  Summoned    := nil;
  Barrier     := nil;
end;

procedure TSummonSpell.Draw;
var
  I: integer;
begin
  for i := 0 to Sources.Count - 1 do
    with TSource(Sources[I]) do
      DrawLine(Summoned.X, Summoned.Y, X, Y, BLACK, ElementCoreColor(Element), Active);
  if Barrier <> nil then
    for I := 0 to Length(Barrier.SavedLines) - 1 do
    begin
      Barrier.SavedLines[I].Draw(Scratch);
    end;
end;

procedure TSummonSpell.Activate;
var
  aX, aY: extended;
  I:      integer;
  P:      array[1..5] of TSource;
  L:      array[1..5] of TLine;
begin
  inherited;
  aX := 0;
  aY := 0;
  for I := 0 to Sources.Count - 1 do
    with TSource(Sources[I]) do
    begin
      aX := aX + X;
      aY := aY + Y;
    end;
  aX := aX / Sources.Count;
  aY := aY / Sources.Count;
  Summoned := SummonClass.Create(Game, aX, aY);

  for I := 1 to 5 do
    P[I] := TSource(Sources[I - 1]);

  L[1] := P[1].Connected(P[4]);
  L[2] := P[4].Connected(P[2]);
  L[3] := P[2].Connected(P[5]);
  L[4] := P[5].Connected(P[3]);
  L[5] := P[3].Connected(P[1]);

  for I := 1 to 5 do
    if L[I] = nil then
      exit;
  Barrier := TAntiDaemonBarrier.Create(Game, Summoned);
  for I := 1 to 5 do
    AddLine(L[I]);
  Barrier.AddLine(L[1], P[1]);
  Barrier.AddLine(L[2], P[4]);
  Barrier.AddLine(L[3], P[2]);
  Barrier.AddLine(L[4], P[5]);
  Barrier.AddLine(L[5], P[3]);
end;

procedure TSummonSpell.Process;
begin
  if Barrier <> nil then
    UseMana(Barrier.TheRate / MANARANGE * 0.5);
  if not UseMana(1) then
  begin
    Summoned.HP := -1;
    if Barrier <> nil then
      Barrier.Disappear := True;
    Free;
    exit;
  end;
end;

{ TGolem }

procedure TGolem.Draw;
begin
  DrawObject(GOLEMLIST, X, Y);
  inherited Draw;
end;

procedure TGolem.Process;
var
  U: TUnit;
  O: TGameObject;
begin
  U := Game.NearestUnit(X, Y, ffGolemFood, Self);

  if WantPlayer or (U = nil) then
    SetDirTo(Game.Player.X, Game.Player.Y)
  else
    SetDirTo(U.X, U.Y);

  if not Move then
    WantPlayer := not WantPlayer;

  if (random < 0.01) or PointsNear(X, Y, Game.Player.X, Game.Player.Y) then
    WantPlayer := False;

  U := Game.NearUnit(X, Y, ffGolemFood, Self);
  if (U <> nil) then
    U.Hit;
  inherited Process;
end;

function TGolem.GetFlags: TFormFlags;
begin
  Result := [ffFight, ffZombieFood];
end;

function TGolem.Armor: Extended;
begin
  Result := 1;
end;

function TGolem.Speed: Extended;
begin
  Result := GOLEMSPEED;
end;

{ TDemon }

procedure TDemon.Draw;
begin
  DrawObject(DEMONLIST, X, Y);
  inherited Draw;
end;

procedure TDemon.Process;
var
  U: TUnit;
begin
  U := Game.NearestUnit(X, Y, ffDemonFood, Self);
  if U <> nil then
    SetDirTo(U.X, U.Y);

  Move;

  U := Game.NearUnit(X, Y, ffDemonFood, Self);
  if U <> nil then
    U.Hit;
  inherited Process;
end;

function TDemon.GetFlags: TFormFlags;
begin
  Result := [ffSeeSources, ffSeeSpells, ffFight]
end;

function TDemon.Armor: Extended;
begin
  Result := 1;
end;

function TDemon.Speed: Extended;
begin
  Result := DEMONSPEED;
end;

{ TChannelSpell }

procedure TChannelSpell.Process;
var
  A, B: TSource;
begin
  inherited Process;
  A := TSource(Sources[0]);
  B := TSource(Sources[1]);
  if (A.MP <= MANACOST) or (B.MP <= MANACOST) then
  begin
    Free;
    exit;
  end;
  if A.MP < B.MP then
  begin
    if B.UseMana(1) then
      A.UseMana(-1)
    else
      Free;
  end
  else if B.MP < A.MP then
  begin
    if A.UseMana(1) then
      B.UseMana(-1)
    else
      Free;
  end;
end;

procedure TChannelSpell.Draw;
var
  A, B: TSource;
begin
  A := TSource(Sources[0]);
  B := TSource(Sources[1]);
  if A.MP < B.MP then
    DrawLine(A.X, A.Y, B.X, B.Y, BLACK, ElementColor(B.Element), Active)
  else if A.MP > B.MP then
    DrawLine(A.X, A.Y, B.X, B.Y, ElementColor(A.Element), BLACK, Active)
  else
    DrawLine(A.X, A.Y, B.X, B.Y, BLACK, BLACK, Active);
end;

constructor TChannelSpell.Create(aGame: TGame; First, Second: TSource);
begin
  inherited Create(aGame);
  AddLine(First.Connected(Second));
  AddLine(First.Connected(Second));
end;

{ TBarrier }

class function TBarrier.PreciseArea: boolean;
begin
  Result := True;
end;

function TBarrier.Callback(U: TUnit; aX, aY: extended): boolean;
begin
  if U.InheritsFrom(TMonster) and (MyArea.Inside(aX, aY) xor
    MyArea.Inside(U.X, U.Y)) then
  begin
    Result := not UseMana(1);
  end
  else
    Result := True;
end;

procedure TBarrier.Process;
begin
  inherited Process;
  if not UseMana(MyArea.TheRate / MANARANGE) then
    Free;
end;

{ TAreaSpell }

constructor TAreaSpell.Create(aGame: TGame);
begin
  inherited;
  MyArea := TSpecialObstracle.Create(aGame, Self);
end;

procedure TAreaSpell.Activate;
var
  I: integer;
  S: TSource;
begin
  inherited;
  if PreciseArea then
  begin
    S := TSource(Sources[0]);
    for I := 0 to Lines.Count - 1 do
    begin
      MyArea.AddLine(TLine(Lines[I]), S);
      S := TLine(Lines[I]).OtherNode(S);
    end;
  end
  else
  begin
    for I := 0 to Sources.Count - 1 do
    begin
      S := TSource(Sources[I]);
      MyArea.AddPoint(S.X, S.Y);
    end;
  end;
end;

procedure TAreaSpell.Draw;
var
  I: integer;
begin
  if not PreciseArea then
    with MyArea do
    begin
      glBegin(GL_POLYGON);
      for I := Low(Points) to High(Points) do
      begin
        SetColor(AreaColor(I / Length(Points)));
        glVertex2f(Points[I].X, Points[I].Y);
      end;
      glEnd;
    end
  else
    inherited;
end;

destructor TAreaSpell.Destroy;
begin
  MyArea.Free;
  inherited;
end;

{ TSpecialObstracle }

constructor TSpecialObstracle.Create(aGame: TGame; aCaller: TAreaSpell);
begin
  inherited Create(aGame);
  Caller := aCaller;
end;

destructor TSpecialObstracle.Destroy;
begin
  SetLength(Points, 0);
  inherited Destroy;
end;

function TPolyObstracle.Inside(aX, aY: extended): boolean;
var
  C, I: integer;
begin
  {Процедура проверки принадлежности точки многоугольнику
//  Версия MY
  }
  C := 0;
  for I := Low(Points) + 1 to High(Points) do
    if IntersectLines(-1000, 0, aX, aY, Points[I - 1].X, Points[I - 1].Y,
      Points[I].X, Points[I].Y) then
      Inc(C);
  if IntersectLines(-1000, 0, aX, aY, Points[High(Points)].X,
    Points[High(Points)].Y, Points[Low(Points)].X, Points[Low(Points)].Y) then
    Inc(C);
  Result := Odd(C);
{  //  Версия TRIANGLE_FAN
  for I := Low(Points)+2 to High(Points) do
    if InTria(Points[0].X, Points[0].Y, Points[I-1].X, Points[I-1].Y, Points[I].X, Points[I].Y) then
    begin
      Result := True;
      exit
    end;
  Result := False;
}
end;

procedure TPolyObstracle.AddPoint(aX, aY: extended);
begin
  if Length(Points) > 0 then
    with Points[High(Points)] do
      if (X = aX) and (Y = aY) then
        exit;

  SetLength(Points, Length(Points) + 1);
  with Points[High(Points)] do
  begin
    X := aX;
    Y := aY;
  end;
  TheRate := RateSize;
end;

procedure TPolyObstracle.AddLine(L: TLine; FromPoint: TSource);
var
  I: integer;
begin
  AddPoint(FromPoint.X, FromPoint.Y);
  if FromPoint = L.Start then
    for I := Low(L.Points) to High(L.Points) do
      with TExtendedPoint(L.Points[I]) do
        AddPoint(X, Y)
  else
    for I := High(L.Points) downto Low(L.Points) do
      with TExtendedPoint(L.Points[I]) do
        AddPoint(X, Y);
  AddPoint(L.OtherNode(FromPoint).X, L.OtherNode(FromPoint).Y);
end;

function TPolyObstracle.CanStep(Who: TUnit; var aX: extended;
  var aY: extended): boolean;
begin
  Result := not Inside(aX, aY);
  {TODO: Добавить проверку пересечений с гранями}
end;

constructor TPolyObstracle.Create(aGame: TGame);
begin
  inherited Create(aGame, 0, 0);
  SetLength(Points, 0);
  TheRate := 0;
end;

function TPolyObstracle.RateSize: extended;
var
  I: integer;
  aMinX, aMaxX, aMinY, aMaxY: extended;
begin
  aMinX := MAXX;
  aMaxX := 0;
  aMinY := MAXY;
  aMaxY := 0;
  for I := 0 to Length(Points) - 1 do
    with Points[I] do
    begin
      if aMaxX < X then
        aMaxX := X;
      if aMinX > X then
        aMinX := X;
      if aMaxY < Y then
        aMaxY := Y;
      if aMinY > Y then
        aMinY := Y;
    end;
  Result := (aMaxX - aMinX) * (aMaxY - aMinY);
end;

function TSpecialObstracle.CanStep(Who: TUnit; var aX: extended;
  var aY: extended): boolean;
begin
  Result := Caller.Callback(Who, aX, aY);
end;

procedure TSpecialObstracle.Process;
var
  I: integer;
  U: TUnit;
begin
  inherited Process;
  for I := Game.Objects.Count - 1 downto 0 do
    if TGameObject(Game.Objects[I]).InheritsFrom(TUnit) then
    begin
      U := TUnit(Game.Objects[I]);
      if Inside(U.X, U.Y) then
        Caller.Callback(U, U.X, U.Y);
    end;
end;

{ THexagram }

function THexagram.CanRepeat: boolean;
begin
  Result := True;
end;

function THexagram.MaxLen: integer;
begin
  Result := 6;
end;

function THexagram.NeedPoly: TPolygonResults;
begin
  Result := [Convex];
end;

function THexagram.Recognize(Game: TGame; Line: TLine): TSpell;
var
  I, J: integer;
  N:    TSource;
  Nodes, Lines: TObjectList;
  L:    TLine;
begin
  Nodes := TObjectList.Create;
  Lines := TObjectList.Create;
  try
    N := FullRecognize(Line, Lines, Nodes);
    if (N <> nil) and (N = Line.Finish) then
    begin
      Result := TSummonSpell.Create(Game, TGolem);
      for I := 0 to Length(Lines.Data) - 1 do
        Result.AddLine(TLine(Lines.Data[I]));
      for I := 0 to Length(Nodes.Data) - 1 do
        for J := 0 to Length(Nodes.Data) - 1 do
        begin
          L := TSource(Nodes.Data[I]).Connected(TSource(Nodes.Data[J]));
          if L <> nil then
            Result.AddLine(L);
        end;
      Result.Activate;
    end
    else
      Result := nil;
  finally
    Nodes.Free;
    Lines.Free;
  end;
end;

function THexagram.CheckLink(Index: integer; Line: TLine; Node: TSource;
  FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;
var
  CX, CY: extended;
begin
  Result := IsFail;
  if (Index > 0) and (Node.Connected(TSource(FinalNodes.PrevNode(2))) = nil) then
    exit
  else if Index = 5 then
    with FinalNodes do
    begin
      if FinalNodes.IsPolygon <> Convex then exit;
      if InTria(X(2), Y(2), X(4), Y(4), X(6), Y(6), X(1), Y(1)) then
        exit;
      if InTria(X(2), Y(2), X(4), Y(4), X(6), Y(6), X(3), Y(3)) then
        exit;
      if InTria(X(2), Y(2), X(4), Y(4), X(6), Y(6), X(5), Y(5)) then
        exit;

      if InTria(X(1), Y(1), X(3), Y(3), X(5), Y(5), X(2), Y(2)) then
        exit;
      if InTria(X(1), Y(1), X(3), Y(3), X(5), Y(5), X(4), Y(4)) then
        exit;
      if InTria(X(1), Y(1), X(3), Y(3), X(5), Y(5), X(6), Y(6)) then
        exit;

      CX := (X(1) + X(3) + X(5)) / 3;
      CY := (Y(1) + Y(3) + Y(5)) / 3;
      if not InTria(X(2), Y(2), X(4), Y(4), X(6), Y(6), CX, CY) then
        exit;
      CX := (X(2) + X(4) + X(6)) / 3;
      CY := (Y(2) + Y(4) + Y(6)) / 3;
      if not InTria(X(1), Y(1), X(3), Y(3), X(5), Y(5), CX, CY) then
        exit;
    end;
  Result := inherited;
end;

{ TSetSequencePattern }

function TSetSequencePattern.CanRepeat: boolean;
begin
  Result := FCanRepeat;
end;

function TSetSequencePattern.MaxLen: integer;
begin
  Result := 10000;
end;

function TSetSequencePattern.NeedPoly: TPolygonResults;
begin
  Result := [Convex]
end;

function TSetSequencePattern.Recognize(Game: TGame; Line: TLine): TSpell;
var
  I: integer;
  N: TSource;
  Nodes, Lines: TObjectList;
begin
  Nodes := TObjectList.Create;
  Lines := TObjectList.Create;
  N     := FullRecognize(Line, Lines, Nodes);
  if (N <> nil) and (N = Line.Finish) then
  begin
    case Style of
      DoDamage: Result := TDamageSpell.Create(Game);
      Heal: Result     := THealSpell.Create(Game);
    end;
    for I := 1 to Length(Nodes.Data) - 1 do
      Result.Sources.Add(Nodes.Data[I]);
    for I := 0 to Length(Lines.Data) - 1 do
      Result.AddLine(TLine(Lines.Data[I]));
    Result.Activate;
  end
  else
    Result := nil;
  Nodes.Free;
  Lines.Free;
end;

constructor TSetSequencePattern.CreateOneElem(Elem: TElement; aStyle: TSpellStyle);
begin
  inherited Create;
  Style      := aStyle;
  FCanRepeat := True;
  Elements   := [];
  Include(Elements, Elem);
end;

constructor TSetSequencePattern.CreateManyElems(Elems: TElements; aStyle: TSpellStyle);
begin
  inherited Create;
  Style      := aStyle;
  FCanRepeat := False;
  Elements   := Elems;
end;

function TSetSequencePattern.CheckLink(Index: integer; Line: TLine;
  Node: TSource; FirstNode: TSource;
  FinalLines, FinalNodes: TObjectList): TLinkCheckResult;
begin
  if Node.Element in Elements then
  begin
    Result := inherited;
    if (Result = PartialOK) and (Node = FirstNode) then
    begin
      if Index > 1 then
        Result := LastOK
      else
        Result := IsFail;
    end;
  end
  else
    Result := IsFail;
end;

{ TObjectList }

constructor TObjectList.Create;
begin
  inherited;
  SetLength(Data, 0);
end;

destructor TObjectList.Destroy;
begin
  SetLength(Data, 0);
  inherited Destroy;
end;

procedure TObjectList.PushObject(S: TObject);
begin
  SetLength(Data, Length(Data) + 1);
  Data[High(Data)] := S;
end;

procedure TObjectList.PopObject;
begin
  SetLength(Data, Length(Data) - 1);
end;

function TObjectList.FindSource(S: TObject): boolean;
var
  I: integer;
begin
  for I := Low(Data) to High(Data) do
    if Data[I] = S then
    begin
      Result := True;
      exit;
    end;
  Result := False;
end;

function TObjectList.PrevNode(N: integer): TObject;
begin
  if N > Length(Data) then
    Result := nil
  else
    Result := Data[High(Data) - N + 1];
end;

function TObjectList.X(I: integer): extended;
begin
  Result := TSource(PrevNode(I)).X;
end;

function TObjectList.Y(I: integer): extended;
begin
  Result := TSource(PrevNode(I)).Y;
end;

function TObjectList.IsPolygon: TPolygonResult;
var
  I, N: integer;
  C:    extended;
  CX, CY: Extended;

function GetCross(X1, Y1, X2, Y2: Extended): Extended;
begin
  Result := X1*Y2-Y1*X2;
end;

function GetCrossC(I, J: Integer): Extended;
begin
  Result := GetCross(X(I)-CX, Y(I)-CY, X(J)-CX, Y(J)-CY);
end;

begin
  Result := CantCheck;
  N      := Length(Data);
  if Data[N-1] = Data[0] then Dec(N);

  if N < 3 then
    exit;
  Result := NotPoly;
  CX := 0;
  CY := 0;
  for I := 1 to N do
  begin
    CX := CX + X(I);
    CY := CY + Y(I);
  end;
  CX := CX/N;
  CY := CY/N;
  C := GetCrossC(N, 1);
  for I := 1 to N-1 do
    if C * GetCrossC(I, I+1) < 0 then
      exit;
  if N = 3 then
    Result := Convex
  else
  begin
    Result := Polygon;
    C := GetCross(X(1)-X(N), Y(1)-Y(N), X(2)-X(1), Y(2)-Y(1));
    for I := 2 to N-1 do
      if C * GetCross(X(I)-X(I-1), Y(I)-Y(I-1), X(I+1)-X(I), Y(I+1)-Y(I)) < 0 then
        exit;
    Result := Convex;
  end;
end;

function TObjectList.LogList: string;
var
  I: Integer;
begin
  Result := '';
  for I := 0 to Length(Data)-1 do
    with TSource(Data[I]) do
      Result := Result + Format('(%d %f,%f) ',[I, X, Y]);
end;

{ TFixedSequencePattern }

function TFixedSequencePattern.CanRepeat: boolean;
begin
  Result := True;
end;

function TFixedSequencePattern.MaxLen: integer;
begin
  Result := Length(Sequence);
end;

function TFixedSequencePattern.NeedPoly: TPolygonResults;
begin
  case Style of
    Channel: Result := inherited NeedPoly;
    Dontknow: Result := [Convex];
  end;
end;

function TFixedSequencePattern.Recognize(Game: TGame; Line: TLine): TSpell;
var
  N: TSource;
  Nodes, Lines: TObjectList;
  I: Integer;
begin
  Nodes := TObjectList.Create;
  Lines := TObjectList.Create;
  N     := FullRecognize(Line, Lines, Nodes);
  if (N <> nil) and (not MustClose or (N = Line.Finish)) then
  begin
    case Style of
      Channel:
      begin
        Result := TChannelSpell.Create(Game, Line.Start, Line.Finish);
        Result.Activate;
      end;
      Dontknow:
      begin
        Result := TDontKnowSpell.Create(Game);
        for I := 1 to Length(Nodes.Data) - 1 do
          Result.Sources.Add(Nodes.Data[I]);
        for I := 0 to Length(Lines.Data) - 1 do
          Result.AddLine(TLine(Lines.Data[I]));
        Result.Activate;
      end;
    end;
  end
  else
    Result := nil;
  Nodes.Free;
  Lines.Free;
end;

constructor TFixedSequencePattern.Create(aSeq: string; aStyle: TSpellStyle;
  aMustClose: boolean);
var
  I, N: integer;
begin
  inherited Create;
  N     := Length(aSeq);
  Style := aStyle;
  MustClose := aMustClose;
  SetLength(Sequence, N);
  for I := 0 to N - 1 do
    case aSeq[I + 1] of
      'F', 'f': Sequence[I] := Fire;
      'W', 'w': Sequence[I] := Water;
      'A', 'a': Sequence[I] := Air;
      'E', 'e': Sequence[I] := Earth;
    end;
end;

function TFixedSequencePattern.CheckLink(Index: integer; Line: TLine;
  Node: TSource; FirstNode: TSource;
  FinalLines, FinalNodes: TObjectList): TLinkCheckResult;
begin
  if Node.Element <> Sequence[Index] then
    Result := IsFail
  else
    Result := inherited;
end;

{ TSequencePattern }

function TSequencePattern.CanRepeat: boolean;
begin
  Result := False;
end;

function TSequencePattern.NeedPoly: TPolygonResults;
begin
  Result := [Convex..CantCheck];
end;

function TSequencePattern.RecognizePart(CurIndex: integer; FromNode: TSource;
  FirstNode: TSource; Exclude, FinalLines, FinalNodes: TObjectList): TSource;
var
  I:    integer;
  L:    TLine;
  Next: TSource;
begin
  {MAIN recursive procedure}
  Result := nil;
  for I := 0 to FromNode.Lines.Count - 1 do
  begin
    L    := TLine(FromNode.Lines[I]);
    Next := L.OtherNode(FromNode);
    if Exclude.FindSource(Next) then
    begin
      Result := nil;
      continue;
    end;
    case CheckLink(CurIndex, L, Next, FirstNode, FinalLines, FinalNodes) of
      PartialOK:
      begin
        if Next <> FirstNode then
        begin
          Exclude.PushObject(FromNode);
          FinalNodes.PushObject(Next);
          FinalLines.PushObject(L);
          Result := RecognizePart(CurIndex + 1, Next, FirstNode, Exclude,
            FinalLines, FinalNodes);
          Exclude.PopObject;
          if Result <> nil then
            exit
          else
          begin
            FinalNodes.PopObject;
            FinalLines.PopObject;
          end;
        end;
      end;
      FullOK:
      begin
        FinalNodes.PushObject(Next);
        FinalLines.PushObject(L);
        if Next <> FirstNode then
        begin
          Exclude.PushObject(FromNode);
          Result := RecognizePart(CurIndex + 1, Next, FirstNode, Exclude,
            FinalLines, FinalNodes);
          Exclude.PopObject;
          if Result = nil then
          begin
            FinalNodes.PopObject;
            FinalLines.PopObject;
          end;
        end;
        if Result = nil then
          Result := Next;
        exit;
      end;
      IsFail:
      begin
        continue;
      end;
      LastOK:
      begin
        FinalNodes.PushObject(Next);
        FinalLines.PushObject(L);
        Result := Next;
        exit;
      end;
    end;
  end;
  Result := nil;
end;

function TSequencePattern.CheckLink(Index: integer; Line: TLine;
  Node: TSource; FirstNode: TSource;
  FinalLines, FinalNodes: TObjectList): TLinkCheckResult;
begin
  if (not CanRepeat) and (Line.Start.Element = Line.Finish.Element) then
    Result := IsFail
  else if not (FinalNodes.IsPolygon in (NeedPoly+[CantCheck])) then
  begin
    Result := IsFail
  end
  else if Index >= MaxLen - 1 then
    Result := LastOK
  else
    Result := PartialOK;
end;

function TSequencePattern.FullRecognize(aLine: TLine;
  FinalLines, FinalNodes: TObjectList): TSource;
var
  L: TObjectList;
begin
  L := TObjectList.Create;
  {Хак: уберем первую линию}
  aLine.Start.Lines.Remove(aLine);
  FinalLines.PushObject(aLine);
  FinalNodes.PushObject(aLine.Finish);
  FinalNodes.PushObject(aLine.Start);

  case CheckLink(0, aLine, aLine.Start, aLine.Finish, FinalLines, FinalNodes) of
    PartialOK:
    begin
      Result := RecognizePart(1, aLine.Start, aLine.Finish, L, FinalLines, FinalNodes);
    end;
    FullOK:
    begin
      Result := RecognizePart(1, aLine.Start, aLine.Finish, L, FinalLines, FinalNodes);
      if Result = nil then
        Result := aLine.Start;
    end;
    IsFail:
      Result := nil;
    LastOK:
      Result := aLine.Start;
  end;
  {Хак: вернем первую линию}
  aLine.Start.Lines.Add(aLine);
  L.Free;

  {Хак: Проверка типа многоугольника}
  if (Result <> nil)then
  begin
    Log(FinalNodes.LogList);
    if not (FinalNodes.IsPolygon in (NeedPoly+[CantCheck]))then
      Result := nil
  end;

end;

{ TDemoSpell }

procedure TDemoSpell.Process;
begin

end;

constructor TDemoSpell.Create(aGame: TGame);
begin
  inherited Create(aGame);
end;

{ TSpell }

constructor TSpell.Create(aGame: TGame);
begin
  inherited Create(aGame, 0, 0);
  Sources := TList.Create;
  Lines   := TList.Create;
end;

procedure TSpell.AddLine(L: TLine);
begin
  L.Start.Lines.Remove(L);
  L.Finish.Lines.Remove(L);
  Lines.Add(L);
  if Sources.IndexOf(L.Start) < 0 then
    Sources.Add(L.Start);
  if Sources.IndexOf(L.Finish) < 0 then
    Sources.Add(L.Finish);
end;

procedure TSpell.Activate;
begin

end;

procedure TSpell.Draw;
var
  I: integer;
begin
  for I := 0 to Lines.Count - 1 do
  begin
    TLine(Lines[I]).Draw(Active);
  end;
end;

procedure TSpell.Process;
begin
  inherited Process;
end;

function TSpell.UseMana(K: extended): boolean;
var
  I: integer;
begin
  Result := True;
  for I := Sources.Count - 1 downto 0 do
    if not TSource(Sources[I]).UseMana(K / Sources.Count) then
    begin
      Result := False;
    end;
end;

destructor TSpell.Destroy;
var
  I: integer;
begin
  Sources.Free;
  for I := Lines.Count - 1 downto 0 do
    TLine(Lines[I]).ToScratch;
  Lines.Free;
  inherited Destroy;
end;

{ TCirclePattern }

function TCirclePattern.Recognize(Game: TGame; Line: TLine): TSpell;
begin
  if Line.Start = Line.Finish then
  begin
    Result := TBarrier.Create(Game);
    Result.AddLine(Line);
    Result.Activate;
  end
  else
    Result := nil;
end;

{ TLine }

constructor TLine.Create(aGame: TGame; aStart: TSource);
begin
  inherited Create;
  Game   := aGame;
  Start  := aStart;
  Finish := nil;
  SetLength(Points, 0);
end;

procedure TLine.AddPoint(aX, aY: extended; Angle: real);
begin
  if abs(Angle - LastAngle) < 1e-3 then
    exit;
  if PointsNear(aX, aY, Start.X, Start.Y) then
    exit;
  LastAngle := Angle;
  SetLength(Points, Length(Points) + 1);
  with Points[High(Points)] do
  begin
    X := aX;
    Y := aY;
  end;
end;

function TLine.Finalize(aFinish: TSource): boolean;
begin
  Result := (Start <> aFinish) or (Length(Points) > 0);
  if not Result then
    exit;
  Finish := aFinish;
  if Finish = nil then
    ToScratch
  else
  begin
    Start.Lines.Add(Self);
    Finish.Lines.Add(Self);
    Game.CheckPatterns(Self);
  end;
end;

procedure TLine.Draw(Style: TLineEffect);
var
  I:      integer;
  FX, FY: extended;
  FirstC, SecondC, MidC, TempC: TOpenGLColor;
begin
  {TODO: LINE_STRIP would be nice}
  FX     := Start.X;
  FY     := Start.Y;
  FirstC := ElementColor(Start.Element);
  if Finish <> nil then
    SecondC := ElementColor(Finish.Element)
  else
    SecondC := FirstC;
  MidC := MixColor(FirstC, SecondC, 1);
  for I := Low(Points) to High(Points) do
    with Points[I] do
    begin
      DrawLine(FX, FY, X, Y, FirstC, MidC, Style);
      TempC  := MidC;
      MidC   := FirstC;
      FirstC := TempC;
      //      MidC := MixColor(MidC, SecondC, 0.2);
      FX     := X;
      FY     := Y;
    end;
  if Finish <> nil then
    DrawLine(FX, FY, Finish.X, Finish.Y, MidC, SecondC, Style)
  else if Self = Game.Player.CurLine then
    DrawLine(FX, FY, Game.Player.X, Game.Player.Y, MidC, SecondC, Style);
end;

function TLine.OtherNode(S: TSource): TSource;
begin
  if S = Start then
    Result := Finish
  else if S = Finish then
    Result := Start
  else
    Result := nil;
end;

procedure TLine.ToScratch;
begin
  //Game.Scratches.Add(Self);{TODO: Убирать ссылки на Start\Finish}
  Free;
end;

{ TSource }

procedure TSource.Draw;
var
  I: integer;
  L: TLine;
begin
  if not (ffSeeSources in Game.Player.GetFlags) then exit;

  glPushMatrix;
  glTranslatef(X, Y, 0);
  Circle(MINDIST / 2, BLACK, TRANSPARENT, BLACK);
  if MP > 0 then
    Circle(MINDIST / 2 * MP, ElementCoreColor(Element), ElementColor(Element), CLR(0, 0, 0, -1));
  glPopMatrix;

  {  C.Pen.Color := clBlack;
  C.Brush.Color := clDkGray;
  C.Polygon([Point(bx-d, by), Point(bx, by-d), Point(bx+d, by), Point(bx, by+d)]);
  d := Round(5*MP);
  C.Brush.Color := ElementColors[Element];
  C.Polygon([Point(bx-d, by), Point(bx, by-d), Point(bx+d, by), Point(bx, by+d)]);
  C.TextOut(bx, by, Name+':'+IntToStr(Round(MP*100))+'%');
}
  for I := 0 to Lines.Count - 1 do
  begin
    L := TLine(Lines[I]);
    if L.Finish = self then
      L.Draw(Ready);
  end;
end;

function TSource.UseMana(K: extended): boolean;
begin
  LastUsed := 10;
  MP     := MP - MANACOST * K;
  Result := MP >= 0;
  if not Result then
    MP := -0.5;
end;

constructor TSource.Create(aGame: TGame; ax, ay, aMP: extended; aElement: TElement);
begin
  inherited Create(aGame, aX, aY);
  MP      := aMP;
  Element := aElement;
  Lines   := TList.Create;
  Rechargeble := True;
end;

destructor TSource.Destroy;
var
  I: integer;
begin
  if (Game.Player.CurLine <> nil) and (Game.Player.CurLine.Start = Self) then
  begin
    Game.Player.CurLine.Finalize(nil);
    Game.Player.CurLine := nil;
  end;
  for I := Lines.Count - 1 downto 0 do
    TLine(Lines[I]).ToScratch;
  Lines.Free;
  inherited;
end;

procedure TSource.Process;
begin
  inherited Process;
  if MP <= 0 then
  begin
    if (Game.Player.CurLine <> nil) and (Game.Player.CurLine.Start = Self) then
    begin
      Game.Player.CurLine.Finalize(nil);
      Game.Player.CurLine := nil;
    end;
    {TODO; memory leak}
    if Rechargeble then
      Lines.Clear
    else
      Game.Objects.Remove(Self);
  end;
  if (LastUsed < 0) and (Lines.Count = 0) and (MP < 1) then
    MP := MP + MANAREGEN
  else
    Dec(LastUsed);
end;

function TSource.Connected(Node: TSource): TLine;
var
  I: integer;
begin
  for I := 0 to Lines.Count - 1 do
    if TLine(Lines[I]).OtherNode(Self) = Node then
    begin
      Result := TLine(Lines[I]);
      exit;
    end;
  Result := nil;
end;

{ TGameObject }

destructor TGameObject.Destroy;
begin
  Broken := True;
  Game.Objects.Remove(Self);
  inherited Destroy;
end;

procedure TGameObject.Draw;
begin

end;

constructor TGameObject.Create(aGame: TGame; aX, aY: extended);
begin
  inherited Create;
  Broken := False;
  Game   := aGame;
  Game.Objects.Add(Self);
  X := aX;
  Y := aY;
end;

procedure TGameObject.Process;
begin

end;

{ TMonster }

procedure TMonster.Draw;
begin
  DrawObject(MONSTERLIST, X, Y);
  inherited;
end;

constructor TMonster.Create(aGame: TGame; aX, aY: extended);
begin
  inherited;
  Sidewalk := False;
end;

procedure TMonster.Process;
var
  U: TUnit;
begin
  U := Game.NearestUnit(X, Y, ffZombieFood, Self);
  if U = nil then
  begin
    Sidewalk := True;
    SetDirTo(X + random * GENOBSTRSIZE * 2 - GENOBSTRSIZE, Y + random * 2 * GENOBSTRSIZE - GENOBSTRSIZE);
  end
  else if not Sidewalk then
    SetDirTo(U.X, U.Y);
  if not Move then
  begin
    Sidewalk := True;
    SetDirTo(X + random * GENOBSTRSIZE * 2 - GENOBSTRSIZE, Y + random * 2 * GENOBSTRSIZE - GENOBSTRSIZE);
  end;

  if Sidewalk and PointsNear(X, Y, TargetX, TargetY) then
    Sidewalk := False;

  U := Game.NearUnit(X, Y, ffZombieFood, Self);
  if U<>nil then
  begin
    Sidewalk := False;
    U.Hit;
  end;
  inherited;
end;

procedure TMonster.Die;
begin
  Dec(Game.MonstersRemains);
  if Game.MonstersRemains <= 0 then
  begin
    ShowMessage('Оставшиеся мертвецы испуганно отступают. Все вокруг залито кровью. ПОБЕДА!');
    Application.Terminate;
  end;
  inherited Die;
end;

function TMonster.GetFlags: TFormFlags;
begin
  Result := [ffBurns, ffFight, ffDemonFood, ffGolemFood];
end;

function TMonster.Armor: Extended;
begin
  Result := 0;
end;

function TMonster.Speed: Extended;
begin
  Result := MONSTERSPEED;
end;

{ TPlayer }

procedure TPlayer.Draw;
var
  phi: GLfloat;
  Tex: Integer;

begin
  //DrawObject(PLAYERLIST, X, Y);
  SetColor(WHITE);
  case CurForm of
  pfWarrior: Tex := TEX_WARRIOR;
  pfZombie:
    begin
      DrawRect(0,0,MAXX, MAXY, ZOMBIESIGHT);
      DrawObject(MONSTERLIST, X, Y);
      Tex := -1;
    end;
  pfDemon:
    begin
      DrawObject(DEMONLIST, X, Y);
      Tex := -1;
    end;
  else
    Tex := TEX_WITCH;
  end;
  if Tex > 0 then
    TexturedRect(X - MINDIST, Y - MINDIST, X + MINDIST, Y + MINDIST, Tex, 0.5);
{  case CurForm of
  if CurForm = pfSpeed then
    TexturedRect(X - MINDIST, Y - MINDIST, X + MINDIST, Y + MINDIST, TEX_SPEED, 0.5);}
  DrawLine(X, Y, TargetX, TargetY, BLACK, BLACK, Target);
  if CurLine <> nil then
    CurLine.Draw(Current);
  inherited;
end;

procedure TPlayer.InsertSource(E: TElement);
begin
  if Inventory[E] > 0 then
  begin
    TSource.Create(Game, X, Y, 1, E).Rechargeble := False;
    Dec(Inventory[E]);
  end;
end;

procedure TPlayer.Die;
begin
  SetScreenGamma(RED);
  UpdateFrame;
  ShowMessage('Game OVER');
  Application.Terminate;
  inherited Destroy;
end;

constructor TPlayer.Create(aGame: TGame; aX, aY: extended);
begin
  inherited;
  DrawMode := False;
  CurLine  := nil;
  Inventory[Fire] := 0;
  Inventory[Water] := 0;
  Inventory[Air] := 0;
  Inventory[Earth] := 0;

  CurForm := pfNormal;
  CurTimeout := 0;
  CurSource := nil;
end;

procedure TPlayer.Process;
var
  U: TUnit;
  I: Integer;
begin
  if CurTimeout > 0 then
  begin
    Dec(CurTimeout);
    if (CurTimeout = 0) or
        ((CurSource <> nil) and not CurSource.UseMana(TRANSFORMCOST)) then
      ReturnForm;
  end;
  CheckLines;
  //Движение
  if not Move then
    DoProcess := False;

  if ffFight in GetFlags then
  begin
    U := Game.NearUnit(X, Y, ffGolemFood, Self);
    if U <> nil then
    begin
      for I := 1 to 3 do U.Hit;
      U.X := U.X - MINDIST*1.5 + Random*MINDIST*3;
      U.Y := U.Y - MINDIST*1.5 + Random*MINDIST*3;
    end;
  end;

  CheckLines;
  inherited Process;
end;

procedure TPlayer.CheckLines;
var
  N: TSource;
begin
  N := TSource(Game.NearObject(X, Y, TSource));

  if not (ffDraw in GetFlags) then
  begin
    if CurLine <> nil then
    begin
      CurLine.AddPoint(X, Y, -1000);
      CurLine.Finalize(nil);
      CurLine := nil;
    end;
    if not (ffSeeSources in GetFlags) and (N<>nil) then
      N.UseMana(10);
    exit
  end;

  //Проверка завершения линии\старта линии
  if CurLine <> nil then
  begin
    if N <> nil then
    begin
      if CurLine.Finalize(N) then
      begin
        CurLine := nil;
        CurLine := TLine.Create(Game, N);
      end;
    end
    else if not DrawMode then
    begin
      CurLine.AddPoint(X, Y, -1000);
      CurLine.Finalize(nil);
      CurLine := nil;
    end
    else
      CurLine.AddPoint(X, Y, arctan2(TargetY - Y, TargetX - X));
  end
  //Проверка
  else if DrawMode and (N <> nil) then
  begin
    CurLine := TLine.Create(Game, N);
  end;
end;

procedure TPlayer.Hit;
begin
  inherited Hit;
  SetScreenGamma(MixColor(RED, TRANSPARENT, 0.6));
end;

function TPlayer.GetFlags: TFormFlags;
begin
  case CurForm of
  pfNormal: Result := [ffSeeSources, ffSeeSpells, ffHeals, ffDraw, ffZombieFood, ffDemonFood];
  pfWarrior: Result := [ffHeals, ffBurns, ffFight, ffZombieFood, ffDemonFood];
  pfZombie: Result := [ffBurns, ffDemonFood, ffGolemFood];
  pfDemon: Result := [ffSeeSources, ffSeeSpells, ffFight, ffZombieFood];
  pfSpeed: Result := [ffSeeSources, ffSeeSpells, ffHeals, ffDraw, ffZombieFood, ffDemonFood];
  pfElemental: Result := [ffSeeSources, ffSeeSpells, ffHeals, ffDraw, ffZombieFood, ffDemonFood];
  end;
end;

procedure TPlayer.SetForm(F: TPlayerForm; Time: Integer; Source: TDontKnowSpell; Elem: TElement);
begin
  SetScreenGamma(MixColor(GREEN, TRANSPARENT, 0.5));
  CurForm := F;
  CurTimeout := Time;
  CurElement := Elem;
  CurSource := Source;
end;

procedure TPlayer.ReturnForm;
begin
  HP := 0.1;
  CurSource := nil;
  CurForm := pfNormal;
end;

function TPlayer.Armor: Extended;
begin
  case CurForm of
  pfWarrior: Result := 1-(1-PLAYERARMOR)/2;
  pfZombie: Result := 0;
  pfDemon: Result := 1-(1-PLAYERARMOR)/2;
  else Result := PLAYERARMOR;
  end;
end;

function TPlayer.Speed: Extended;
begin
  case CurForm of
  pfWarrior: Result := PLAYERSPEED;
  pfZombie: Result := MONSTERSPEED;
  pfDemon: Result := DEMONSPEED;
  pfSpeed: Result := PLAYERSPEED*2;
  else Result := PLAYERSPEED;
  end;
end;

{ TRectObstracle }

constructor TRectObstracle.Create(aGame: TGame; ax1, ay1, ax2, ay2: extended);
begin
  inherited Create(aGame, ax1, ay1);
  X2 := ax2;
  Y2 := ay2;
end;

procedure TRectObstracle.Draw;
begin
  SetColor(GRAY);
  TexturedRect(X, Y, X2, Y2, TEX_WALL, 1);
end;

function TRectObstracle.CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean;
begin
  Result := not (InRange(aX, X, X2) and InRange(aY, Y, Y2));
end;

{ TGame }

procedure TGame.InitPatterns;
begin
  //Barrier
  Patterns.Add(TCirclePattern.Create);
  //Channel

  Patterns.Add(TFixedSequencePattern.Create('FF', Channel, True));
  Patterns.Add(TFixedSequencePattern.Create('EE', Channel, True));
  Patterns.Add(TFixedSequencePattern.Create('WW', Channel, True));
  Patterns.Add(TFixedSequencePattern.Create('AA', Channel, True));

  //Don't know
  Patterns.Add(TFixedSequencePattern.Create('FAWE', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('FEWA', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('WAFE', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('WEFA', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('AFEW', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('AWEF', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('EWAF', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('EFAW', Dontknow, True));

  //Damage
    Patterns.Add(TSetSequencePattern.CreateOneElem(Fire, DoDamage));


 //Summon
  Patterns.Add(THexagram.Create);
  Patterns.Add(TPentagram.Create);

  //Heal
  Patterns.Add(TSetSequencePattern.CreateManyElems([Earth, Water, Air], Heal));

end;

function TGame.CheckPatterns(L: TLine): TSpell;
var
  I: integer;
begin
  Result := nil;
  for I := 0 to Patterns.Count - 1 do
  begin
    Result := TPattern(Patterns[I]).Recognize(Self, L);
    if Result <> nil then
      exit;
  end;
end;

procedure TGame.ProcessTime;
var
  I:    integer;
  O:    TGameObject;
  X, Y: extended;
begin
  I := 0;
  for I := Objects.Count - 1 downto 0 do
  begin
    if I < Objects.Count then
    begin
      O := TGameObject(Objects[I]);
      O.Process;
    end;
  end;
  if random < CurSpawnRate then
  begin
    case random(4) of
      0:
      begin
        X := 0;
        Y := Random * MAXY;
      end;
      1:
      begin
        X := MAXX;
        Y := Random * MAXY;
      end;
      2:
      begin
        X := Random * MAXX;
        Y := 0;
      end;
      3:
      begin
        X := Random * MAXX;
        Y := MAXY;
      end;
    end;
    TMonster.Create(Self, X, Y);
  end;
  CurSpawnRate := CurSpawnRate + SPAWNINC;
end;

procedure TGame.Clear;
var
  I: integer;
begin
{  for I := Objects.Count - 1 downto 0 do
    if TGameObject(Objects[I]).InheritsFrom(TSpell) then
    TGameObject(Objects[I]).Free;}{TODO: MEMORY LEAK!}
  Objects.Clear;

  for I := Scratches.Count - 1 downto 0 do
    TLine(Scratches[I]).Free;
  Scratches.Clear;
  Player    := nil;
  VIEWSCALE := DEFVIEWSCALE;
  CurSpawnRate := INITSPAWNRATE;
  MonstersRemains := TARGETMONSTERS;
end;

procedure TGame.ClearTransTable;
var
  E1, E2: TElement;
  P1, P2: TPassType;
begin
  for E1 := Low(TElement) to High(TElement) do
    for E2 := Low(TElement) to High(TElement) do
      for P1 := Low(TPassType) to High(TPassType) do
        for P2 := Low(TPassType) to High(TPassType) do
          TRANSTABLE[E1, E2, P1, P2] := teNothing;
end;

procedure TGame.GenTransTable;
var
  E1, E2: TElement;
  P1, P2: TPassType;
  MaxRate, R: Integer;
  T: TTransformEffect;
begin
  MaxRate := 0;
  for T := Low(TTransformEffect) to High(TTransformEffect) do
    Inc(MaxRate, TransformRates[T]);
  for E1 := Low(TElement) to High(TElement) do
    for E2 := Low(TElement) to High(TElement) do
      for P1 := Low(TPassType) to High(TPassType) do
        for P2 := Low(TPassType) to High(TPassType) do
        begin
          R := Random(MaxRate);
          for T := Low(TTransformEffect) to High(TTransformEffect) do
          begin
            Dec(R, TransformRates[T]);
            if R <=0 then
            begin
              TRANSTABLE[E1, E2, P1, P2] := T;
              break
            end
          end;
        end;
end;

procedure TGame.LoadLevel(Filename: string);
var
  ff: TextFile;
  X, Y, X2, Y2, I, N: integer;
begin
  Clear;
  CurSpawnRate := 0;
  AssignFile(ff, Filename);
  Reset(ff);
  ReadLn(ff, X, Y);
  Player := TPlayer.Create(Self, X, Y);
  ReadLn(ff, Player.Inventory[Fire], Player.Inventory[Water],
    Player.Inventory[Air], Player.Inventory[Earth]);
  ReadLn(ff, N);
  for I := 1 to N do
  begin
    ReadLn(ff, X, Y);
    TDemon.Create(Self, X, Y);
  end;
  ReadLn(ff, N);
  for I := 1 to N do
  begin
    ReadLn(ff, X, Y);
    TMonster.Create(Self, X, Y);
  end;
  ReadLn(ff, N);
  for I := 1 to N do
  begin
    ReadLn(ff, X, Y, X2, Y2);
    TRectObstracle.Create(Self, X, Y, X2, Y2);
  end;
  ReadLn(ff, N);
  for I := 1 to N do
  begin
    ReadLn(ff, X, Y, X2);
    TSource.Create(Self, X, Y, 1, TElement(X2));
  end;
  CloseFile(ff);
end;

procedure TGame.GenLevel;
var
  I:   integer;
  X, Y, A, B, C, D: extended;
  Obs: TObstracle;
begin
  CurSpawnRate := INITSPAWNRATE;
  Clear;
  Player := TPlayer.Create(Self, MAXX / 2, MAXY / 2);
  Player.Inventory[Fire] := 10;
  Player.Inventory[Earth] := 10;
  Player.Inventory[Water] := 10;
  Player.Inventory[Air] := 10;
  for I := 1 to GENOBSTR do
  begin
    case Random(3) of
      0:
      begin
        A   := random * GENOBSTRSIZE / 2 + GENOBSTRSIZE / 2;
        B   := random * GENOBSTRSIZE / 2 + GENOBSTRSIZE / 2;
        X   := Random * (MAXX-A*1.1);
        Y   := Random * (MAXY-B*1.1);
        Obs := TRectObstracle.Create(Self, X, Y, X + A, Y + B);
      end;
      1:
      begin
        A   := random * GENOBSTRSIZE / 4 + GENOBSTRSIZE / 10;
        X   := Random * (MAXX-2.5*A)+A*1.25;
        Y   := Random * (MAXY-2.5*A)+A*1.25;
        Obs := TRoundObstracle.Create(Self, X, Y, A);
      end;
      2:
      begin
        A   := random * GENOBSTRSIZE / 4 + GENOBSTRSIZE / 4;
        B   := random * GENOBSTRSIZE / 4 + GENOBSTRSIZE / 4;
        C   := random * GENOBSTRSIZE / 4 + GENOBSTRSIZE / 4;
        D   := random * GENOBSTRSIZE / 4 + GENOBSTRSIZE / 4;
        X   := Random * (MAXX-A*1.1-C*1.1)+A*1.1;
        Y   := Random * (MAXY-C*1.1-D*1.1);
        Obs := TTriObstracle.Create(Self, X, Y, X - A, Y + B, X + C, Y + D);
      end;
    end;
    if not Obs.CanStep(Player, Player.X, Player.Y) then
      Obs.Free;
  end;
  for I := 1 to GENSOURCES do
  begin
    repeat
      X := Random * MAXX;
      Y := Random * MAXY;
    until NearObject(X, Y, TSource, nil) = nil;
    TSource.Create(Self, X, Y, 1, RandomElement);
  end;
  SetScreenGamma(YELLOW);
end;

procedure TGame.Draw;
var
  I: integer;
begin
  for I := 0 to Scratches.Count - 1 do
    TLine(Scratches[I]).Draw(Scratch);
  if ffSeeSpells in Player.GetFlags then
    for I := Objects.Count - 1 downto 0 do
      with TGameObject(Objects[I]) do
        if InheritsFrom(TSpell) then
          Draw;
  for I := Objects.Count - 1 downto 0 do
    with TGameObject(Objects[I]) do
      if not InheritsFrom(TSpell) then
        Draw;
end;

function TGame.CheckObstracles(Who: TUnit; var X: extended; var Y: extended): boolean;
var
  I: integer;
begin
  if not (InRange(X, 0, MAXX) and InRange(Y, 0, MAXY)) then
  begin
    Result := False;
    exit;
  end;
  Result := True;
  for I := 0 to Objects.Count - 1 do
  begin
    if TGameObject(Objects[I]).InheritsFrom(TObstracle) and not
      TObstracle(Objects[I]).CanStep(Who, X, Y) then
    begin
      Result := False;
    end;
  end;
end;

constructor TGame.Create;
begin
  inherited;
  VIEWSCALE := DEFVIEWSCALE;
  CurSpawnRate := INITSPAWNRATE;
  Objects := TList.Create;
  Scratches := TList.Create;
  Patterns := TList.Create;
  Player := TPlayer.Create(Self, MAXX / 2, MAXY / 2);
  InitPatterns;
end;

function TGame.NearObject(X, Y: Extended; aClass: TGameClass;
  Exclude: TGameObject = nil): TGameObject;
var
  I: integer;
  O: TGameObject;
begin
  for I := 0 to Objects.Count - 1 do
  begin
    O := TGameObject(Objects[I]);
    if O.InheritsFrom(aClass) and (O <> Exclude) and PointsNear(X, Y, O.X, O.Y) then
    begin
      Result := O;
      exit;
    end;
  end;
  Result := nil;
end;

function TGame.NearUnit(X, Y: Extended; ForFlag: TFormFlag; Exclude: TGameObject
  ): TUnit;
var
  I: integer;
  O: TGameObject;
begin
  for I := 0 to Objects.Count - 1 do
  begin
    O := TGameObject(Objects[I]);
    if O.InheritsFrom(TUnit) and (O <> Exclude) and PointsNear(X, Y, O.X, O.Y) and (ForFlag in TUnit(O).GetFlags) then
    begin
      Result := TUnit(O);
      exit;
    end;
  end;
  Result := nil;
end;

function TGame.NearestObject(X, Y: Extended; aClass: TGameClass;
  Exclude: TGameObject): TGameObject;
var
  I: integer;
  O: TGameObject;
  MinRange, D: extended;
begin
  Result   := nil;
  MinRange := Sqr(MAXX + MAXY);
  for I := 0 to Objects.Count - 1 do
  begin
    O := TGameObject(Objects[I]);
    if (not O.InheritsFrom(aClass)) or (O = Exclude) then
      continue;
    D := Sqr(X - O.X) + Sqr(Y - O.Y);
    if D < MinRange then
    begin
      MinRange := D;
      Result   := O;
    end;
  end;
end;

function TGame.NearestUnit(X, Y: Extended; ForFlag: TFormFlag; Exclude: TGameObject
  ): TUnit;
var
  I: integer;
  O: TGameObject;
  MinRange, D: extended;
begin
  Result   := nil;
  MinRange := Sqr(MAXX + MAXY);
  for I := 0 to Objects.Count - 1 do
  begin
    O := TGameObject(Objects[I]);
    if (not O.InheritsFrom(TUnit)) or (O = Exclude) or not (ForFlag in TUnit(O).GetFlags) then
      continue;
    D := Sqr(X - O.X) + Sqr(Y - O.Y);
    if D < MinRange then
    begin
      MinRange := D;
      Result   := TUnit(O);
    end;
  end;
end;

{ TUnit }

function TUnit.Move: boolean;
var
  R, aX, aY: extended;
begin
  R := sqr(X - TargetX) + sqr(Y - TargetY);
  if R < sqr(Speed) then
  begin
    aX := TargetX;
    aY := TargetY;
  end
  else
  begin
    aX := X + Speed / sqrt(R) * (TargetX - X);
    aY := Y + Speed / sqrt(R) * (TargetY - Y);
  end;
  Result := Game.CheckObstracles(Self, aX, aY);
  if Result then
  begin
    X := aX;
    Y := aY;
  end;
end;

procedure TUnit.Hit;
begin
  if random < armor then
    exit;
  HP := HP - random * DAMAGE;
end;

procedure TUnit.Die;
begin
  Destroy;
end;

procedure TUnit.SetDirTo(ToX, ToY: extended);
begin
  TargetX := ToX;
  TargetY := ToY;
end;

procedure TUnit.Draw;
var
  x1, x2, xr, y1, y2: extended;
begin
  X1 := X - 10;
  X2 := X + 10;
  Y1 := Y - 10;
  Y2 := Y - 8;
  XR := (X2 - X1) * HP + X1;
  DrawRect(XR, Y1, X2, Y2, BLACK);
  DrawRect(X1, Y1, XR, Y2, GREEN);
end;

procedure TUnit.Process;
begin
  inherited;
  if HP <= 0 then
    Die;
end;

constructor TUnit.Create(aGame: TGame; aX, aY: extended);
begin
  inherited;
  TargetX := aX;
  TargetY := aY;
  HP      := 1;
end;

procedure TUnit.Heal;
begin
  HP := min(1, HP + HEALPOWER);
end;

end.

